use std::{collections::HashSet, sync::Arc};

use axum::http::Uri;
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use tokio::sync::{RwLock, Semaphore};
use tokio_util::sync::CancellationToken;

use crate::repo::Repo;

pub type AppStateShared<R> = Arc<RwLock<AppState<R>>>;

pub struct AppState<R: Repo> {
    pub repo: R,
    pub url_handling_semaphore: Semaphore,
    pub cancellation_token: CancellationToken,
}

impl<R: Repo> AppState<R> {
    pub fn new(url_handlers_pool_size: usize, repo: R, ct: CancellationToken) -> Self {
        Self {
            repo,
            url_handling_semaphore: Semaphore::new(url_handlers_pool_size),
            cancellation_token: ct,
        }
    }

    pub fn regen_cancellation_token(&mut self) {
        self.cancellation_token = CancellationToken::new();
    }
}

#[derive(Serialize)]
pub struct UrlResult {
    pub status: Option<u16>,
    #[serde(serialize_with = "serialize_uri")]
    pub url: Uri,
    pub content_length: usize,
    pub error_msg: Option<String>,
}

fn serialize_uri<S: Serializer>(uri: &Uri, serializer: S) -> Result<S::Ok, S::Error> {
    serializer.serialize_str(&uri.to_string())
}

#[derive(Serialize)]
pub struct UrlReportResponse {
    pub successes: Vec<UrlResult>,
    pub failures: Vec<UrlResult>,
}

#[derive(Deserialize)]
pub struct UrlRequest {
    #[serde(deserialize_with = "deser_uri")]
    pub urls: HashSet<Uri>,
}

fn deser_uri<'de, D>(deserializer: D) -> Result<HashSet<Uri>, D::Error>
where
    D: Deserializer<'de>,
{
    let raw_urls = HashSet::<String>::deserialize(deserializer)?;
    raw_urls
        .into_iter()
        .map(|url| {
            Uri::try_from(&url)
                .map_err(|e| serde::de::Error::custom(format!("{}, url: '{url}'", e.to_string())))
        })
        .collect::<Result<HashSet<_>, _>>()
}
