use std::collections::{HashMap, HashSet};

use sled::{Batch, Db};

use crate::{
    error::{AppError, AppResult},
    models::UrlResult,
};

pub trait Repo: Send + Sync + 'static {
    fn insert_urls(&self, urls: &[UrlResult]) -> AppResult<()>;

    fn delete_urls(&self, urls: Option<&HashSet<String>>) -> AppResult<usize>;

    fn get_urls(&self, urls: Option<&HashSet<String>>) -> AppResult<HashMap<String, usize>>;
}

impl Repo for Db {
    fn insert_urls(&self, urls: &[UrlResult]) -> AppResult<()> {
        let mut batch = Batch::default();
        for url in urls {
            batch.insert(
                url.url.to_string().as_bytes(),
                bincode::serialize(&url.content_length)?,
            );
        }
        self.apply_batch(batch).map_err(AppError::from)
    }

    fn delete_urls(&self, urls: Option<&HashSet<String>>) -> AppResult<usize> {
        let mut deleted = 0;

        match urls {
            Some(urls) => {
                for url in urls {
                    info!("deleting {url}");
                    let del = self.remove(url.as_bytes())?;
                    if del.is_some() {
                        deleted += 1
                    }
                }
            }
            None => {
                info!("deleting all urls");
                deleted = self.len();
                self.clear()?;
            }
        };

        Ok(deleted)
    }

    fn get_urls(&self, urls: Option<&HashSet<String>>) -> AppResult<HashMap<String, usize>> {
        match urls {
            Some(urls) => {
                let mut result = HashMap::new();

                for url in urls {
                    let Some(val) = self.get(url.as_bytes())? else {
                        continue;
                    };
                    result.insert(url.to_owned(), bincode::deserialize(&val)?);
                }
                Ok(result)
            }
            None => self
                .iter()
                .map(|result| {
                    let (k, v) = result?;
                    let key = String::from_utf8(k.to_vec())?;
                    let val = bincode::deserialize(&v)?;
                    Ok((key, val))
                })
                .collect::<AppResult<_>>(),
        }
    }
}
